/*
 * query.c: for rcpt500 (X.500 email query responder)
 *
 * 18 June 1992 by Mark C Smith
 * Copyright (c) 1992 The Regents of The University of Michigan
 * All Rights Reserved
 */

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <ctype.h>

#include <lber.h>
#include <ldap.h>
#include "rcpt500.h"

extern int dosyslog;
extern int strip;
extern int derefaliases;
extern int sizelimit;
extern char *ldaphost;
extern char *searchbase;
extern char *dapuser;

static char buf[ MAXSIZE ];
static char *errpreface = "Your query failed: ";


int
query_cmd( msgp, reply )
    struct msginfo	*msgp;
    char		*reply;
{
    LDAP		*ldp;
    LDAPMessage		*ldmsgp;
    char		*s, *dn;
    int			spaces, opid, matches, approximate;
    char		filter[ 1024 ];

    if ( msgp->msg_arg == NULL ) {
	return( help_cmd( msgp, reply ));
    }

    remove_trailing_space( msgp->msg_arg );
    if ( *msgp->msg_arg == '\0' ) {
	return( help_cmd( msgp, reply ));
    }

    /*
     * turn '.'s, '_'s and '*'s into spaces
     */
    for ( spaces = 0, s = msgp->msg_arg; *s != '\0'; ++s ) {
	if ( *s == '.' || *s == '_' || *s == '*' ) {
	    *s = ' ';
	    ++spaces;
	} else if ( *s == ' ' ) {
	    ++spaces;
	}
    }

    /*
     * open connection to LDAP server and bind as dapuser
     */
    if (( ldp = ldap_open( ldaphost, 0 )) == NULL ) {
	strcat( reply, errpreface );
	strcat( reply, "X.500 service unavailable" );
	return( 0 );
    }

    if ( ldap_bind( ldp, dapuser, NULL, LDAP_AUTH_SIMPLE ) != LDAP_SUCCESS ) {
	report_ldap_err( ldp, reply );
	ldap_unbind( ldp );
	return( 0 );
    }

    /*
     * set options for search and build filter
     */
    ldp->ld_deref = derefaliases;
    ldp->ld_sizelimit = sizelimit;

    approximate = 0;
    if ( spaces > 0 ) {
	sprintf( filter, "(|(cn=%s)(sn=%s))", msgp->msg_arg, msgp->msg_arg );
    } else {
	sprintf( filter, "(|(cn=%s)(sn=%s)(uid=%s))", msgp->msg_arg,
		msgp->msg_arg, msgp->msg_arg );
    }
    
    if (( matches = do_search( ldp, filter, &ldmsgp )) < 0 ) {
	report_ldap_err( ldp, reply );
	ldap_unbind( ldp );
	return( 0 );
    }

    if ( matches == 0 ) {
	approximate = 1;
	if ( spaces ) {
	    sprintf( filter, "(|(cn~=%s)(sn~=%s))", msgp->msg_arg,
		    msgp->msg_arg );
	} else {
	    sprintf( filter, "(sn~=%s)", msgp->msg_arg );
	}
	if (( matches = do_search( ldp, filter, &ldmsgp )) < 0 ) {
	    report_ldap_err( ldp, reply );
	    ldap_unbind( ldp );
	    return( 0 );
	}
    }

    if ( matches == 0 ) {
	sprintf( buf, "No matches were found for '%s'\n", msgp->msg_arg );
	strcat( reply, buf );
	ldap_unbind( ldp );
	return( 0 );
    }

    if ( ldp->ld_errno == LDAP_TIMELIMIT_EXCEEDED
	    || ldp->ld_errno == LDAP_SIZELIMIT_EXCEEDED ) {
	strcat( reply, "(Partial results only - a limit was exceeded)\n" );
    }

    if ( matches == 1 ) {
	if ( approximate ) {
	    sprintf( buf, "One approximate match was found '%s':\n",
		    msgp->msg_arg );
	} else {
	    sprintf( buf, "One exact match was found for '%s':\n",
		    msgp->msg_arg );
	}
	strcat( reply, buf );
	dn = ldap_get_dn( ldp, ldap_first_entry( ldp, ldmsgp ));
	if ( do_read( ldp, dn, reply ) < 0 ) {
	    free( dn );
	    report_ldap_err( ldp, reply );
	    ldap_unbind( ldp );
	    return( 0 );
	}
	free( dn );
	
    } else {
	if ( approximate ) {
	    sprintf( buf, "%d approximate matches were found for '%s':\n",
		    matches, msgp->msg_arg );
	} else {
	    sprintf( buf, "%d matches were found for '%s':\n",
		    matches, msgp->msg_arg );
	}
	strcat( reply, buf );
	append_entry_list( reply, msgp->msg_arg, ldp, ldmsgp );
    }

    ldap_unbind( ldp );
    return( 0 );
}


int
do_search( ldp, filter, ldmpp )
    LDAP	*ldp;
    char	*filter;
    LDAPMessage	**ldmpp;
{
    int		opid, err;
    static char	*attrs[] = { "cn", "title", NULL };

    if (( opid = ldap_search( ldp, searchbase, LDAP_SCOPE_SUBTREE, filter,
		attrs, 0 )) == -1 ) {
	return( -1 );
    }

    if ( ldap_result( ldp, opid, 1, NULL, ldmpp ) == -1 ) {
	return( -1 );
    }

    err = ldap_result2error( ldp, *ldmpp, 0 );
    if ( err == LDAP_SUCCESS || err == LDAP_TIMELIMIT_EXCEEDED
	    || err == LDAP_SIZELIMIT_EXCEEDED ) {
	return( ldap_count_entries( ldp, *ldmpp ));
    } else {
	return( 0 );
    }
}


append_entry_list( reply, query, ldp, ldmsgp )
    char	*reply;
    char	*query;
    LDAP	*ldp;
    LDAPMessage	*ldmsgp;
{
    LDAPMessage	*e;
    char	*dn, *rdn, *s, **title;
    int		free_rdn = 0;

    for ( e = ldap_first_entry( ldp, ldmsgp ); e != NULL;
		e = ldap_next_entry( ldp, e )) {
	dn = ldap_get_dn( ldp, e );
	if (( s = strchr( dn, ',' )) != NULL ) {
	    *s = '\0';
	}
	if (( s = strchr( dn, '=' )) == NULL ) {
	    rdn = dn;
	} else {
	    rdn = s + 1;
	}

	/*
	 * if this entry's rdn is an exact match for the thing looked up, we
	 * return the CN that has a digit after it, so that the user is
	 * returned something guaranteed to yield exactly one match if they
	 * pick it from the list and query it
	 */

	if ( stricmp( rdn, query ) == 0 ) {
	    char	**cn;
	    int		i;

	    if (( cn = ldap_get_values( ldp, e, "cn" )) != NULL ) {
		for ( i = 0; cn[i] != NULL; i++ ) {
		    if ( isdigit( *( cn[i] + strlen( cn[i] ) - 1 ))) {
			rdn = strdup( cn[i] );
			free_rdn = 1;
			break;
		    }
		}
		ldap_value_free( title );
	    }
	}

	title = ldap_get_values( ldp, e, "title" );
	sprintf( buf, "  %-20s    %s\n", rdn, title ? title[0] : "" );
	strcat( reply, buf );
	if ( title != NULL ) {
	    ldap_value_free( title );
	}
	free( dn );
	if ( free_rdn ) {
	    free( rdn );
	}
    }
}


static
append_attr( reply, label, attr, e, ldp, multiline )
    char	*reply;
    char	*label;
    char	*attr;
    LDAPMessage	*e;
    LDAP	*ldp;
    int		multiline;
{
    char	**val;
    int		i, gotone = 0;

    if ( (val = ldap_get_values( ldp, e, attr )) == NULL ) {
	if ( strcmp( attr, "mail" ) == 0 ) {
	    sprintf( buf, "%-19s\n", label );
	    strcat( reply, buf );
	    sprintf( buf,
		    "                    None registered in this service.\n" );
	    strcat( reply, buf );
	}
	return;
    }

    sprintf( buf, "%-19s\n", label );
    strcat( reply, buf );
    for ( i = 0; val[i] != NULL; i++ ) {
	if ( multiline ) {
	    char	*s, *p;

	    if ( gotone ) {
		sprintf( buf, "%-19s\n", label );
		strcat( reply, buf );
	    }

	    p = s = val[i];
	    while ( s = strchr( s, '$' ) ) {
		*s++ = '\0';
		while ( isspace( *s ) )
			s++;
		sprintf( buf, "                    %s\n", p );
		strcat( reply, buf );
		p = s;
	    }
	    sprintf( buf, "                    %s\n", p );
	    strcat( reply, buf );
	    gotone = 1;
	} else {
	    sprintf( buf, "                    %s\n", val[i] );
	    strcat( reply, buf );
	}
    }

    ldap_value_free( val );
}


int
do_read( ldp, dn, reply )
    LDAP	*ldp;
    char	*dn;
    char	*reply;
{
    int		opid;
    LDAPMessage	*ldmp, *entry;
    short	i;
    char	*s, *ufn;
    static char	*attrs[] = { "cn", "postalAddress", "uid", "title",
			     "mail", "telephoneNumber",
			     "facsimileTelephoneNumber", 
			     "multiLineDescription", NULL };

    if (( opid = ldap_search( ldp, dn, LDAP_SCOPE_BASE, "objectclass=*",
		attrs, 0 )) == -1 ) {
	return( -1 );
    }

    if ( ldap_result( ldp, opid, 1, NULL, &ldmp ) == -1 ) {
	return( -1 );
    }

    if ( ldap_result2error( ldp, ldmp, 0 ) != LDAP_SUCCESS ) {
	return( -1 );
    }

    entry = ldap_first_entry( ldp, ldmp );
    dn = ldap_get_dn( ldp, entry );
    ufn = ldap_dn2ufn( dn );
    free( dn );
    for ( i = 0; i < strip; i++ ) {
	if ( (s = strrchr( ufn, ',' )) == NULL )
		break;
	*s = '\0';
    }
    sprintf( buf, "\"%s\"\n", ufn );
    strcat( reply, buf );
    free( ufn );

    append_attr( reply, "  Also known as:", "cn", entry, ldp, 0 );
    append_attr( reply, "  E-mail address:", "mail", entry, ldp, 0 );
    append_attr( reply, "  Fax number:", "facsimileTelephoneNumber",
	    entry, ldp, 0 );
    append_attr( reply, "  Business phone:", "telephoneNumber", entry, ldp, 0 );
    append_attr( reply, "  Business address:", "postalAddress", entry, ldp, 1 );
    append_attr( reply, "  Title:", "title", entry, ldp, 0 );
    append_attr( reply, "  Uniqname:", "uid", entry, ldp, 0 );
    append_attr( reply, "  Description:", "multiLineDescription",
	    entry, ldp, 1 );
    ldap_msgfree( ldmp );
    return( 0 );
}


report_ldap_err( ldp, reply )
    LDAP	*ldp;
    char	*reply;
{
    strcat( reply, errpreface );
    strcat( reply, ldap_err2string( ldp->ld_errno ));
    strcat( reply, "\n" );
}


remove_trailing_space( s )
    char	*s;
{
    char	*p = s + strlen( s ) - 1;

    while ( isspace( *p ) && p > s ) {
	--p;
    }
    *(++p) = '\0';
}
